/*
 * $FreeBSD: src/tools/KSE/ksetest/kse_threads_test.c,v 1.1 2002/06/29 17:39:07 julian Exp $
 */
/* #define _KERNEL */

#include <assert.h>
#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <ucontext.h>
#include <unistd.h>
#include <sys/kse.h>
#include <sys/queue.h>

/*#define DEBUG*/
/*************************************************************
 * These should probably be in a .h file
 **************************************************************/
typedef void thread_fn(void *arg);

struct user_thread {
	struct thread_mailbox mbox;
	ucontext_t	uc;
	TAILQ_ENTRY(user_thread) runq_next;
	char		*name;
	int		done;
};

struct per_kse {
	struct kse_mailbox mbox;
	struct user_thread *curthread;
	ucontext_t	uc;
	int		initialized;
};
/*************************************************************
 * Debug stuff
 **************************************************************/
#define UTS_CONTEXT
#ifdef UTS_CONTEXT
#define SETCONTEXT(arg)	thr_setcontext(arg, NULL, 0)
#else
#define SETCONTEXT(arg)	thr_setcontext(arg, 0)
#endif
#ifndef	DEBUG
#define dbg_log(args...)
#define DUMPREGS(desc) do {} while (0)
#else
#define dbg_log(args...)	printf(args, ##args)

jmp_buf jb3;
#define DUMPREGS(desc) do {_setjmp(jb3); printjb(jb3, desc); } while (0)

char *regname[] = {"%eip","%ebx","%esp","%ebp",
			"%esi","%edi","fpcr","MSK0",
			"MSK1","MSK2","MSK3"};


static 
printjb(struct _jmp_buf *jb, char *desc)
{
	
	int i;
	printf("jb (%s) is at 0x%x\n", desc, jb);
	for( i = 0; i< _JBLEN-4; i++) {
		printf("jb[%d] (%s) = 0x%x\n", i, regname[i], jb[0]._jb[i]);
	}
}
#endif

/*************************************************************
 * Globals
 **************************************************************/
struct per_kse first_kse; /* for NOW cheat and make it global */
struct per_kse *current_kse;
TAILQ_HEAD(, user_thread) runqueue = TAILQ_HEAD_INITIALIZER(runqueue);
static sigset_t	process_mask;
static ucontext_t main_uc;
volatile int	stop_all = 0;
volatile int	stop_5 = 0;
volatile int	stop_6 = 0;

/*************************************************************
 * Implementation parameters
 **************************************************************/
#define T_STACKSIZE	(16*4096)	/* thread stacksize */
#define K_STACKSIZE	(1*4096)	/* KSE (UTS) stacksize */

/*************************************************************
 * UTS funcions.
 * Simple round_robin for now.
 **************************************************************/
static void
runq_insert(struct user_thread *thread)
{
	TAILQ_INSERT_TAIL(&runqueue, thread, runq_next);
}

static struct user_thread *
select_thread(void)
{
	struct user_thread *thread;

	if ((thread = TAILQ_FIRST(&runqueue))) {
		TAILQ_REMOVE(&runqueue, thread, runq_next);
	}
	return (thread);
}

/*
 * For now just emulate a way to get and set the current kse and thread.
 */
static struct per_kse *
get_curkse()
{
	return (current_kse);
}

static void
set_curkse(struct per_kse *kse)
{
	current_kse = kse;
}

static struct user_thread *
get_curthread()
{
	return ((get_curkse())->curthread);
}

static void
set_curthread(struct user_thread *thread)
{
	(get_curkse())->curthread = thread;
}

/*************************************************************
 * The UTS upcall entrypoint
 * Called once on startup (and left by longjump)
 * and there-after, returned to by the upcall many times.
 **************************************************************/
void
UTS(struct kse_mailbox *ke_mbox)
{
	struct user_thread *thread;
	struct thread_mailbox *completed;
	struct per_kse *ksedata;
	int ret;

	/**********************************/
	/* UTS upcall starts running here. */
	/**********************************/
	/**********************************/

	ksedata = ke_mbox->kmbx_UTS_handle;

	/* Make sure this upcall doesn't get interrupted? */
	ke_mbox->kmbx_current_thread = NULL;

	if (ksedata->initialized == 0) {
		/* first time initialization */
		current_kse = ksedata;
		ksedata->initialized = 1;
		/*
		 * XXX - For i386, load %gs appropriately so we can get
		 * to per-kse data (outside of the UTS upcall).
		 */
		dbg_log("UTS should load gs here.\n");
		/*
		 * Setup our context so we can jump here from userland
		 * as well as getting an upcall from kernel.
		 */
		assert(thr_getcontext(&ksedata->uc) == 0);
	}
	if (ksedata->curthread != NULL) {
		/* See if the current thread needs any work. */
		if (ksedata->curthread->done != 0) {
			/*
			 * This thread is done; free it's resources
			 * and wakeup any thread that is waiting on it.
			 */
			dbg_log("UTS found completed thread.\n");
		}
	}


	/* If there are returned syscall threads, put them on the run queue */
	if ((completed = ke_mbox->kmbx_completed_threads)) {
		ke_mbox->kmbx_completed_threads = NULL;
		while (completed) {
			thread = completed->UTS_handle;
			dbg_log("UTS found thread %s in mailbox\n",
			    thread->name);
			/*
			 * For now, just copy the saved mcontext into a
			 * ucontext so we can use setcontext to resume
			 * the thread.
			 */
			memcpy(&thread->uc.uc_mcontext,
			    &thread->mbox.ctx.mcontext,
			    sizeof(thread->uc.uc_mcontext));
			thread->uc.uc_mcontext.mc_flags = __UC_MC_VALID;
			completed = completed->next_completed;
			runq_insert(thread);
		}
	}

	/* find highest priority thread and load it */
	if ((thread = select_thread())) {
		dbg_log("UTS about to start thread %s\n", thread->name);
		ksedata->curthread = thread;
		ret = thr_setcontext(&thread->uc,
		    &ke_mbox->kmbx_current_thread, &thread->mbox);
		assert(ret == 0);
		/* NOTREACHED */
	}
	kse_yield(); 		/* in the kernel it does a thread_exit() */
	/* NOTREACHED */
} 

/*************************************************************
 * Startup mechanism functions
 **************************************************************/
static int
kickkse(struct per_kse *ksedata, int newgroup)
{
	char * newstack;
	struct kse_mailbox *mboxaddr;
	int err;

	/*
	 * Save the current kse.  In the future we will use a
	 * reserved register to store this information (%gs on
	 * i386).
	 */
	current_kse = ksedata;

	newstack = malloc(K_STACKSIZE);
	mboxaddr = &ksedata->mbox;
	mboxaddr->kmbx_stackbase = newstack;
	mboxaddr->kmbx_stacksize = K_STACKSIZE;
	mboxaddr->kmbx_upcall = UTS;
	mboxaddr->kmbx_UTS_handle = ksedata;
	err = kse_new(mboxaddr, newgroup);
	return(err);
}


static int
startkse(struct per_kse *ksedata)
{
	return (kickkse(ksedata, 0));
}

static int
startksegrp(struct per_kse *ksedata)
{
	return(kickkse(ksedata, 1));
}

static void
thread_yield()
{
	struct per_kse *kse;
	volatile int once;

	if (((kse = get_curkse()) != NULL) && (kse->curthread != NULL)) {
		/* Enter a critical region: */
		kse->mbox.kmbx_current_thread = NULL;

		/*
		 * Save the current threads context, insert it into
		 * the run queue, and select a new thread.
		 */
		once = 0;
		thr_getcontext(&kse->curthread->uc);
		if (once == 0) {
			once = 1;
			runq_insert(kse->curthread);
			dbg_log("Thread '%s' yielding to ",
			    kse->curthread->name);
			kse->curthread = select_thread();
			dbg_log("thread '%s'\n", kse->curthread->name);
			thr_setcontext(&kse->curthread->uc,
			    &kse->mbox.kmbx_current_thread,
			    &kse->curthread->mbox);
		}
	}
}

static void
thread_wrapper(void *(*func) (void *), int arg1, void *arg2)
{
	void *ret;
	struct per_kse *kse;

	if ((kse = get_curkse()) != NULL)
		dbg_log("Starting thread %s\n", kse->curthread->name);
	/* We only pass arg2 for now. */
	ret = func(arg2);
	dbg_log("Thread finishing.\n");

	/*
	 * Essentially enter a critical region.  This will prevent
	 * the kernel from making another upcall to this KSE.
	 */
	kse = get_curkse();
	kse->mbox.kmbx_current_thread = NULL;

	dbg_log("Thread %s has left the building.\n", kse->curthread->name);
	/*
	 * Mark this thread as being completed so any per-thread
	 * resources can be freed and any threads wanting to join
	 * to it can be notified.
	 */
	kse->curthread->done = 1;

	/*
	 * Go back to the UTS scheduler.  The thread just completed
	 * will be cleaned there.
	 */
	SETCONTEXT(&kse->uc);
	/* Never returns. */
}

struct user_thread *
makethread(thread_fn *fn, char *name, int arg1, void *arg2)
{
	struct user_thread *tcb;

	dbg_log("Making thread %s\n", name);

	/* We could combine these mallocs */
	tcb = malloc(sizeof *tcb);
	bzero(tcb, sizeof(*tcb));
	tcb->mbox.UTS_handle = tcb; /* back pointer */
	if (name != NULL)
		tcb->name = strdup(name);
	else
		tcb->name = strdup("unnamed");

	/* malloc the thread's stack */
	/* We COULD mmap it with STACK characteristics */
	/* Then we could add a guard page. */
	tcb->uc.uc_stack.ss_size = T_STACKSIZE; /* set the size we want */
	tcb->uc.uc_stack.ss_sp = malloc(T_STACKSIZE);
	tcb->uc.uc_link = NULL;
	tcb->uc.uc_sigmask = process_mask;

	/*
	 * The context passed in to makecontext() is required to be
	 * an already valid context.
	 */
	assert(thr_getcontext(&tcb->uc) == 0);
	makecontext(&tcb->uc, (void (*)())thread_wrapper, 3, fn, arg1, arg2);
	return (tcb);
}

/*************************************************************
 * code for three separate threads. (so we can see if it works)
 *************************************************************/
static void 
thread1_code(void *arg)
{
	dbg_log("Thread 1 started.\n");
	while (stop_all == 0) {
		sleep (1);
		write(1,".",1);
		thread_yield();
	}
}

static void 
thread2_code(void *arg)
{
	dbg_log("Thread 2 started.\n");
	while (stop_all == 0) {
		sleep (2);
		write(1,"+",1);
		thread_yield();
	}
}

static void 
thread3_code(void *arg)
{
	int i;

	dbg_log("Thread 3 started.\n");
	for (i = 0; i < 10; i++) {
		sleep (5);
		thread_yield();
		write(1,"=",1);
		if (i == 5) {
			/* Stop threads 5 and 6. */
			stop_5 = 1;
			stop_6 = 1;
		}
	}
	/* Stop everyone else. */
	stop_all = 1;
}

static void
thread4_code(void *arg)
{
	dbg_log("Thread 4 started.\n");
	SETCONTEXT(&main_uc);
}

static void 
thread5_code(void *arg)
{
	int counts = 0;

	dbg_log("Thread 5 started.\n");
	while (stop_5 == 0) {
		counts++;
		if ((counts % 10000) == 0) {
			sleep(1);
			write(1,"\\",1);
		} else {
			thread_yield();
			(void)getpid();
		}
	}
}

static void 
thread6_code(void *arg)
{
	int counts = 0;

	dbg_log("Thread 6 started.\n");
	while (stop_6 == 0) {
		counts++;
		if ((counts % 10000) == 0) {
			sleep(1);
			write(1,"/",1);
		} else {
			thread_yield();
			(void)getpid();
		}
	}
}

int main()
{
	struct user_thread *td4 = NULL;
	volatile int once = 0;

	/* set up global structures */
	TAILQ_INIT(&runqueue);
	sigprocmask(SIG_SETMASK, NULL, &process_mask);

	/* First make sure normal context switching works: */
	thr_getcontext(&main_uc);
	if (once == 0) {
		once = 1;
		td4 = makethread(thread4_code, "thread 4", 0, NULL);
		dbg_log ("Thread 4 was made, ptr 0x%p\n", td4);
		SETCONTEXT(&td4->uc);
	}
	else {
		dbg_log("Returned from thread 4.\n");
	}

	/* define two threads to run, they are runnable but not yet running */
	runq_insert(makethread(thread1_code, "thread 1", 0, NULL));
	runq_insert(makethread(thread2_code, "thread 2", 0, NULL));
	runq_insert(makethread(thread5_code, "thread 5", 0, NULL));
	runq_insert(makethread(thread6_code, "thread 6", 0, NULL));

	/* and one which we will run ourself */
	first_kse.curthread = makethread(&thread3_code, "thread 3", 0, NULL);
	first_kse.initialized = 0;

	/* start two KSEs in different KSEGRPs */
	if (startkse(&first_kse)) {
		perror("failed to start KSE");
		exit(1);
	}

	/* startksegrp(&second_kse); */ /* we can't do 2 KSEs yet */
	/* One will be sufficient */

	/*
	 * We are a thread, start the ball rolling and let the
	 * kernel know we are it.  This will set the current threads
	 * mailbox in the kse which will allow the kernel to give
	 * us an upcall when the thread blocks (actually, any
	 * system call for now).
	 */
	thr_setcontext(&first_kse.curthread->uc,
	    &first_kse.mbox.kmbx_current_thread, &first_kse.curthread->mbox);

	return (0);
}

